# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Analogue and digital filter design utilities.

This module is mainly a wrapper around `scipy.signal` filter design,
with convenience functions for common filter types, and support for
LIGO-specific filter design conventions.
"""

from __future__ import annotations

import operator
from functools import reduce
from math import (
    log10,
    pi,
)
from typing import (
    TYPE_CHECKING,
    cast,
    overload,
)

import numpy
from astropy.units import (
    Quantity,
    Unit,
)
from numpy import fft as npfft
from scipy import signal
from scipy.signal.windows import tukey

from .window import get_window

# filter type definitions
if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        Literal,
        TypeAlias,
        TypeVar,
    )

    from astropy.units import UnitBase
    from astropy.units.typing import QuantityLike
    from numpy.typing import (
        ArrayLike,
        NDArray,
    )

    from ..typing import (
        Array1D,
        ArrayLike1D,
    )
    from .window import WindowLike

    TArray = TypeVar("TArray", bound=Array1D)

    LinearTimeInvariant: TypeAlias = signal.lti | signal.dlti

    # FIR
    TapsType = Array1D
    # IIR
    FilterTypeName: TypeAlias = Literal["butter", "cheby1", "cheby2", "ellip"]
    SosType: TypeAlias = numpy.ndarray[tuple[int, Literal[6]]]
    ZpkCompatible: TypeAlias = tuple[ArrayLike1D, ArrayLike1D, float]
    ZpkType: TypeAlias = tuple[Array1D, Array1D, float]
    BAType: TypeAlias = tuple[Array1D, Array1D]
    IirFilterCompatible: TypeAlias = SosType | ZpkCompatible | BAType
    IirFilterType: TypeAlias = SosType | ZpkType | BAType
    # generic
    FilterCompatible: TypeAlias = TapsType | IirFilterCompatible | LinearTimeInvariant
    FilterType: TypeAlias = TapsType | IirFilterType | LinearTimeInvariant

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__all__ = [
    "bandpass",
    "concatenate_zpks",
    "fir_from_transfer",
    "frequency_response",
    "highpass",
    "is_sos",
    "is_zpk",
    "lowpass",
    "notch",
    "parse_filter",
    "prepare_analog_filter",
    "prepare_digital_filter",
]

HERTZ = Unit("Hz")
RAD_S = Unit("rad/s")

TWO_PI: float = 2 * pi

# All ZPK filters have three components
ZPK_NARGS = 3

# All SOS filters are 2D arrays with six coefficients per section
SOS_NDIM = 2
SOS_NCOEFF = 6


# -- Utilities -----------------------

def _as_float(
    x: QuantityLike,
    unit: str | UnitBase = "Hz",
) -> float:
    """Convert input to a float in the given units."""
    return Quantity(x, unit).value


def _is_hertz(unit: str | UnitBase) -> bool:
    """Return `True` if ``unit`` represents Hertz."""
    try:
        return Unit(unit) == HERTZ
    except ValueError:
        return False


def _is_radians(unit: str | UnitBase) -> bool:
    """Return `True` if ``unit`` represents radians per second or sample."""
    # rad/sample isn't an astropy Unit, so check string manually
    if unit in ("rad/s", "rad/sample"):
        return True
    try:
        return Unit(unit) == RAD_S
    except ValueError:
        return False


def is_zpk(zpktup: object) -> bool:
    """Return `True` if ``zpktup`` looks like a ZPK-format filter definition.

    Returns
    -------
    iszpk : `bool`
        `True` if input argument looks like a 3-tuple giving arrays of
        zeros and poles, and a gain (`float`).
    """
    return (
        # Input object is a length 3 tuple/list
        isinstance(zpktup, tuple | list)
        and len(zpktup) == ZPK_NARGS
        # First two elements are 1D arrays/lists/tuples
        and isinstance(zpktup[0], list | tuple | numpy.ndarray)
        and numpy.ndim(cast("ArrayLike", zpktup[0])) == 1
        and isinstance(zpktup[1], list | tuple | numpy.ndarray)
        and numpy.ndim(cast("ArrayLike", zpktup[1])) == 1
        # Third element is a float
        and isinstance(zpktup[2], float)
    )


def is_sos(sos: object) -> bool:
    """Return `True` if ``sos`` looks like a SOS-format filter definition.

    Returns
    -------
    issos : `bool`
        `True` if input argument looks like a 2D-array giving second-order
        sections.
    """
    return (
        isinstance(sos, numpy.ndarray)
        and sos.ndim == SOS_NDIM
        and sos.shape[1] == SOS_NCOEFF
    )


def concatenate_zpks(*zpks: ZpkCompatible) -> ZpkType:
    """Concatenate a list of zero-pole-gain (ZPK) filters.

    Parameters
    ----------
    *zpks
        One or more zero-pole-gain format, each one should be a 3-`tuple`
        containing an array of zeros, an array of poles, and a gain `float`.

    Returns
    -------
    zeros : `numpy.ndarray`
        The concatenated array of zeros.
    poles : `numpy.ndarray`
        The concatenated array of poles.
    gain : `float`
        The overall gain.

    Examples
    --------
    Create a lowpass and a highpass filter, and combine them:

    >>> from gwpy.signal.filter_design import (
    ...     highpass, lowpass, concatenate_zpks)
    >>> hp = highpass(100, 4096)
    >>> lp = lowpass(1000, 4096)
    >>> zpk = concatenate_zpks(hp, lp)

    Plot the filter:

    >>> from gwpy.plot import BodePlot
    >>> plot = BodePlot(zpk, sample_rate=4096)
    >>> plot.show()
    """
    zeros, poles, gains = zip(*zpks, strict=True)
    return (
        numpy.concatenate(zeros),
        numpy.concatenate(poles),
        reduce(operator.mul, gains, 1),
    )


def num_taps(
    sample_rate: float,
    transitionwidth: float,
    gpass: float,
    gstop: float,
) -> int:
    """Return the number of taps for an FIR filter with the given shape.

    Parameters
    ----------
    sample_rate : `float`
        Sampling rate of target data.

    transitionwidth : `float`
        The width (in the same units as `sample_rate`) of the transition
        from stop-band to pass-band.

    gpass : `float`
        The maximum loss in the passband (dB).

    gstop : `float`
        The minimum attenuation in the stopband (dB).

    Returns
    -------
    numtaps : `int`
       The number of taps for an FIR filter.

    Notes
    -----
    Credit: http://dsp.stackexchange.com/a/31077/8223
    """
    gpass = 10 ** (-gpass / 10.)
    gstop = 10 ** (-gstop / 10.)
    ntaps = int(
        2 / 3.
        * log10(1 / (10 * gpass * gstop))
        * sample_rate
        / abs(transitionwidth),
    )
    # highpass filters must have an odd number of taps
    if transitionwidth > 0 and ntaps % 2 == 0:
        return ntaps + 1
    return ntaps


# -- Formatting / conversions --------

@overload
def _transform(
    filt: BAType | ZpkType | SosType,
    from_: Literal["zpk", "ba", "sos"],
    to_: Literal["zpk"],
) -> ZpkType: ...
@overload
def _transform(
    filt: BAType | ZpkType | SosType,
    from_: Literal["zpk", "ba", "sos"],
    to_: Literal["ba"],
) -> BAType: ...
@overload
def _transform(
    filt: BAType | ZpkType | SosType,
    from_: Literal["zpk", "ba", "sos"],
    to_: Literal["sos"],
) -> SosType: ...

def _transform(  # noqa: PLR0911
    filt: BAType | ZpkType | SosType,
    from_: Literal["zpk", "ba", "sos"],
    to_: Literal["zpk", "ba", "sos"],
    **kwargs,
) -> BAType | ZpkType | SosType:
    """Transform a filter between different representations.

    Parameters
    ----------
    filt : `tuple` or `numpy.ndarray`
        Filter definition in one of the supported formats.

    from_ : `str`
        Format of input filter definition.
        One of: ``'zpk'``, ``'ba'``, ``'sos'``.

    to_ : `str`
        Format of output filter definition.
        One of: ``'zpk'``, ``'ba'``, ``'sos'``.

    kwargs
        Additional keyword arguments passed to the underlying
        `scipy.signal` conversion functions.

    Returns
    -------
    filt_converted : `tuple` or `numpy.ndarray`
        Filter definition in the requested format.

    Raises
    ------
    ValueError
        If `from_` or `to_` are not one of the supported formats.

    See Also
    --------
    scipy.signal.sos2tf
    scipy.signal.sos2zpk
    scipy.signal.tf2sos
    scipy.signal.tf2zpk
    scipy.signal.zpk2sos
    scipy.signal.zpk2tf
    """
    # Null transformation
    if from_ == to_:
        return filt

    if to_ == "zpk" and from_ == "ba":
        filt = cast("BAType", filt)
        return signal.tf2zpk(*filt, **kwargs)

    if to_ == "zpk" and from_ == "sos":
        filt = cast("SosType", filt)
        return signal.sos2zpk(filt, **kwargs)

    if to_ == "ba" and from_ == "zpk":
        filt = cast("ZpkType", filt)
        return signal.zpk2tf(*filt, **kwargs)

    if to_ == "ba" and from_ == "sos":
        filt = cast("SosType", filt)
        return signal.sos2tf(filt, **kwargs)

    if to_ == "sos" and from_ == "zpk":
        filt = cast("ZpkType", filt)
        return signal.zpk2sos(*filt, **kwargs)

    if to_ == "sos" and from_ == "ba":
        filt = cast("BAType", filt)
        return signal.tf2sos(*filt, **kwargs)

    msg = f"unknown filter transformation: '{from_}' -> '{to_}'"
    raise ValueError(msg)


def _normalize_gain_hz_to_rad(
    zeros: Array1D,
    poles: Array1D,
    gain: float,
) -> float:
    """Normalize gain when converting Hz to rad/s.

    This preserves the frequency response magnitude by scaling the gain
    according to: k' = k * |∏p_i / ∏z_i| * (2π)^(n_p - n_z)

    Parameters
    ----------
    zeros : `numpy.ndarray`
        Zeros in rad/s (already converted from Hz).

    poles : `numpy.ndarray`
        Poles in rad/s (already converted from Hz).

    gain : `float`
        Original gain (from Hz specification).

    Returns
    -------
    normalized_gain : `float`
        Gain scaled to preserve frequency response.
    """
    # Compute product of non-zero poles and zeros
    nonzero_poles = poles[poles != 0]
    nonzero_zeros = zeros[zeros != 0]
    if len(nonzero_zeros) > 0 and len(nonzero_poles) > 0:
        # Scale by ratio of pole to zero magnitudes
        gain *= numpy.abs(numpy.prod(nonzero_poles) / numpy.prod(nonzero_zeros))

    # Account for poles/zeros at the origin
    n_poles_at_origin = poles.size - nonzero_poles.size
    n_zeros_at_origin = zeros.size - nonzero_zeros.size

    # Poles at origin: multiply by (2π)^n_p
    # Zeros at origin: divide by (2π)^n_z
    if n_poles_at_origin > 0 or n_zeros_at_origin > 0:
        gain *= TWO_PI ** (n_poles_at_origin - n_zeros_at_origin)

    return gain


def _convert_zpk_units(
    zpk: ZpkCompatible,
    unit: str | UnitBase,
    *,
    normalize_gain: bool = False,
) -> ZpkType:
    """Convert an analogue ZPK between unit conventions.

    Parameters
    ----------
    zpk : `tuple`
        (zeros, poles, gain) tuple.

    unit : `str`
        The units in which the zeros and poles are specified.
        Either ``'Hz'`` or ``'rad/s'``.

    normalize_gain : `bool`, optional
        Whether to normalize the gain when converting from Hz to rad/s.

        - `False` (default):
          Multiply zeros/poles by -2π but leave gain unchanged.
          This matches the LIGO GDS **'f' plane** convention
          (``plane='f'`` in ``s2z()``).

        - `True`:
          Normalize gain to preserve frequency response magnitude.
          Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
          Use this when your filter was designed with the transfer
          function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
          This matches the LIGO GDS **'n' plane** convention
          (``plane='n'`` in ``s2z()``).

        Ignored when `unit='rad/s'`.

    Returns
    -------
    zeros : `numpy.ndarray` of `numpy.cfloat`
        Converted zeros in rad/s.

    poles : `numpy.ndarray` of `numpy.cfloat`
        Converted poles in rad/s.

    gain : `float`
        Adjusted gain accounting for unit conversion and normalization.

    Warns
    -----
    UserWarning
        If `normalize_gain` is set when `unit='rad/s'` (normalization is ignored).

    Raises
    ------
    ValueError
        If `unit` is not 'Hz' or 'rad/s'.

    Notes
    -----
    This function converts TO rad/s (s-plane). For conversions to other
    representations, use `scipy.signal.bilinear_zpk()` for z-plane, or
    `scipy.signal.zpk2tf()` for transfer function form.
    """
    zeros, poles, gain = zpk
    zeros = numpy.array(zeros, dtype=numpy.complex128)
    poles = numpy.array(poles, dtype=numpy.complex128)

    if _is_hertz(unit):
        # Convert frequencies from Hz to rad/s
        for zi in range(len(zeros)):
            zeros[zi] *= -TWO_PI
        for pj in range(len(poles)):
            poles[pj] *= -TWO_PI

        # Apply gain normalization if requested
        if normalize_gain:
            gain = _normalize_gain_hz_to_rad(zeros, poles, gain)

    elif _is_radians(unit):
        # Already in correct units, warn if normalize_gain was set
        if normalize_gain:
            import warnings
            warnings.warn(
                f"normalize_gain parameter is ignored when unit='{unit}' "
                "(filter is already in rad/s)",
                UserWarning,
                stacklevel=2,
            )
    else:
        msg = (
            "zpk can only be given with unit='Hz', 'rad/s', or 'rad/sample', "
            f"not '{unit}'"
        )
        raise ValueError(msg)

    return zeros, poles, gain


@overload
def _convert_to_digital(
    filt: SosType | ZpkType | LinearTimeInvariant,
    sample_rate: float,
) -> tuple[Literal["zpk"], ZpkType]: ...
@overload
def _convert_to_digital(
    filt: TapsType | BAType,
    sample_rate: float,
) -> tuple[Literal["ba"], BAType]: ...

def _convert_to_digital(
    filt: FilterType,
    sample_rate: float,
) -> tuple[Literal["ba", "zpk"], IirFilterType]:
    """Convert an analog filter to digital via bilinear functions.

    Parameters
    ----------
    filt: `tuple`
        Input filter to convert.

    sample_rate: `float`
        Sample rate of digital data that will be filtered.

    Returns
    -------
    dform : `str`
        Type of filter.

    dfilter : 'tuple'
        Digital filter values.
    """
    form, filt_ = parse_filter(filt)

    if form == "ba":
        filt_ = cast("BAType", filt_)
        return form, signal.bilinear(*filt_, fs=sample_rate)
    if form == "zpk":
        filt_ = cast("ZpkType", filt_)
        return form, signal.bilinear_zpk(*filt_, fs=sample_rate)

    msg = f"cannot convert '{form}' to digital"
    raise ValueError(msg)


@overload
def parse_filter(
    filt: TapsType | BAType | signal.TransferFunction,
) -> tuple[Literal["ba"], BAType]: ...
@overload
def parse_filter(
    filt: SosType,
) -> tuple[Literal["sos"], SosType]: ...
@overload
def parse_filter(
    filt: ZpkCompatible | LinearTimeInvariant,
) -> tuple[Literal["zpk"], ZpkType]: ...

def parse_filter(
    filt: FilterCompatible,
) -> tuple[Literal["ba", "sos", "zpk"], BAType | ZpkType | SosType]:
    """Parse arbitrary input args into a TF, ZPK, or SOS filter definition.

    No transformations are applied to the filter, it is simply unpacked
    into a standard form.

    Parameters
    ----------
    filt : `numpy.ndarray` or `tuple`
        Filter definition.
        Any of the following formats are supported:

        - FIR taps as a 1D `numpy.ndarray`
        - IIR transfer function as a 2-tuple of 1D `numpy.ndarray`
        - IIR zero-pole-gain as a 3-tuple of 1D `numpy.ndarray` and a `float`
        - IIR second-order-sections as a 2D `numpy.ndarray`
        - `scipy.signal.lti` or `scipy.signal.dlti` object

    Returns
    -------
    ftype : `str`
        Either ``'ba'`` ``'zpk'`` or ``'sos'``.
    filt : `numpy.ndarray` or `tuple`
        The filter components for the returned ``ftype``.
        If ``'ba'``, a 2-tuple of numerator and denominator arrays.
        If ``'zpk'``, a 3-tuple of zeros, poles, and gain.
        If ``'sos'``, a 2D array of second-order sections.
    """
    # Parse SOS filter
    if is_sos(filt):  # sos
        return "sos", cast("SosType", filt)

    # Parse FIR taps
    if isinstance(filt, numpy.ndarray) and filt.ndim == 1:  # fir
        b, a = cast("numpy.ndarray[tuple[int]]", filt), numpy.ones(1)
        return "ba", (b, a)

    # Use lti to parse the rest
    if isinstance(filt, signal.lti | signal.dlti):
        lti_ = filt
    else:
        lti_ = signal.lti(*filt)  # ty: ignore[invalid-argument-type]
    if isinstance(lti_, signal.TransferFunction):
        return "ba", (lti_.num, lti_.den)
    if not isinstance(lti_, signal.ZerosPolesGain):
        lti_ = lti_.to_zpk()      # ty: ignore[invalid-argument-type]
    return "zpk", (lti_.zeros, lti_.poles, lti_.gain)


def _prewarp_zpk(
    zeros: Array1D,
    poles: Array1D,
    gain: float,
    fs: float,
) -> tuple[Array1D, Array1D, float]:
    """Apply prewarping to analogue ZPK before bilinear transform.

    This implements the GDS prewarping step (iirutil.cc lines 154-158):
    For each pole/zero at frequency ω (rad/s):
        mag = |ω|
        g = (2*fs / mag) * tan(mag / (2*fs))
        ω_warped = ω * g
        gain *= g  (for zeros: divide, for poles: multiply)

    Parameters
    ----------
    zeros : `numpy.ndarray`
        Zeros in rad/s.

    poles : `numpy.ndarray`
        Poles in rad/s.

    gain : `float`
        System gain.

    fs : `float`
        Sampling frequency in Hz.

    Returns
    -------
    zeros_warped : `numpy.ndarray`
        Prewarped zeros.

    poles_warped : `numpy.ndarray`
        Prewarped poles.

    gain_warped : `float`
        Adjusted gain accounting for prewarping.

    Notes
    -----
    This prewarping step ensures that after the bilinear transform,
    the digital filter's frequency response matches the analog design
    at the specified pole/zero frequencies. This is the default behavior
    in LIGO GDS (fPrewarp = true).
    """
    zeros_warped = zeros.copy()
    poles_warped = poles.copy()
    gain_warped = gain

    fs2 = 2.0 * fs

    # Prewarp zeros
    for i in range(len(zeros)):
        mag = numpy.abs(zeros[i])
        if mag > 0:
            g = (fs2 / mag) * numpy.tan(mag / fs2)
            zeros_warped[i] = zeros[i] * g
            gain_warped /= g  # Gain adjustment for zeros

    # Prewarp poles
    for i in range(len(poles)):
        mag = numpy.abs(poles[i])
        if mag > 0:
            g = (fs2 / mag) * numpy.tan(mag / fs2)
            poles_warped[i] = poles[i] * g
            gain_warped *= g  # Gain adjustment for poles

    return zeros_warped, poles_warped, gain_warped


def prepare_analog_filter(
    filt: FilterCompatible,
    *,
    unit: str | UnitBase = "Hz",
    normalize_gain: bool = False,
) -> tuple[Literal["ba", "zpk", "sos"], BAType | ZpkType | SosType]:
    """Prepare an analog filter by parsing and converting units.

    This handles:

    1. Parsing filter specification
    2. Converting Hz → rad/s for ZPK filters (if unit='Hz')
    3. Applying gain normalization (if requested)

    Does NOT:

    - Convert to digital
    - Extract gain for SOS
    - Apply prewarping

    Parameters
    ----------
    filt : filter specification
        Filter as ``(b, a)``, ``(z, p, k)``, sos array,
        or `~scipy.signal.lti` object.

    unit : `str`, optional
        For analogue ZPK filters, the units in which the zeros and poles are
        specified. Either ``'Hz'`` or ``'rad/s'`` (default).

    normalize_gain : `bool`, optional
        Whether to normalize the gain when converting from Hz to rad/s.

        - `False` (default):
          Multiply zeros/poles by -2π but leave gain unchanged.
          This matches the LIGO GDS **'f' plane** convention
          (``plane='f'`` in ``s2z()``).

        - `True`:
          Normalize gain to preserve frequency response magnitude.
          Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
          Use this when your filter was designed with the transfer
          function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
          This matches the LIGO GDS **'n' plane** convention
          (``plane='n'`` in ``s2z()``).

        Ignored when `unit='rad/s'`.

    Returns
    -------
    form : `str`
        Filter form: ``'ba'``, ``'zpk'``, or ``'sos'``.

    filt : `tuple` or `numpy.ndarray`
        Filter coefficients in the identified form.

    See Also
    --------
    prepare_digital_filter
        Prepare filter for digital filtering with prewarping and gain extraction.

    parse_filter
        Parse filter specification without unit conversion.
    """
    form, filter_tuple = parse_filter(filt)

    if form == "zpk" and _is_hertz(unit):
        filter_tuple = cast("ZpkType", filter_tuple)
        filter_tuple = _convert_zpk_units(
            filter_tuple,
            unit=unit,
            normalize_gain=normalize_gain,
        )

    return form, filter_tuple


@overload
def prepare_digital_filter(
    filt: FilterCompatible,
    *,
    analog: bool = ...,
    sample_rate: QuantityLike = ...,
    unit: str | UnitBase = ...,
    normalize_gain: bool = ...,
    output: Literal["ba"] = ...,
) -> BAType: ...
@overload
def prepare_digital_filter(
    filt: FilterCompatible,
    *,
    analog: bool = ...,
    sample_rate: QuantityLike = ...,
    unit: str | UnitBase = ...,
    normalize_gain: bool = ...,
    output: Literal["zpk"] = ...,
) -> ZpkType: ...
@overload
def prepare_digital_filter(
    filt: FilterCompatible,
    *,
    analog: bool = ...,
    sample_rate: QuantityLike = ...,
    unit: str | UnitBase = ...,
    normalize_gain: bool = ...,
    output: Literal["sos"] = ...,
) -> SosType: ...

def prepare_digital_filter(
    filt: FilterCompatible,
    *,
    analog: bool = False,
    sample_rate: QuantityLike = 1.0,
    unit: str | UnitBase = "Hz",
    normalize_gain: bool = False,
    prewarp: bool = True,
    output: Literal["zpk", "ba", "sos"] = "zpk",
) -> BAType | ZpkType | SosType:
    """Prepare a filter for digital filtering.

    This function parses the input filter specification, optionally converts
    an analog filter to digital using prewarping + bilinear transform,
    and returns the filter in the requested format.

    For incoming digital filters, this function basically does nothing.

    Parameters
    ----------
    filt : filter specification
        Filter as ``(b, a)``, ``(z, p, k)``, sos array,
        or `~scipy.signal.lti` object.

        For digital filters (``analog=False``), the input filter coefficients
        are assumed to already be in digital form (z-domain).

        For analog filters (``analog=True``), the input filter coefficients
        should be in the units specified by ``unit`` (default: Hz).

    analog : `bool`, optional
        When `True`, the input filter is analog and will be converted to
        digital using prewarping + bilinear transform (GDS method).
        When `False` (default), the input filter is already digital.

    sample_rate : `float`, `~astropy.units.Quantity`, optional
        Sampling frequency (Hz) of the digital system.
        Required when ``analog=True``.

    unit : `str`, optional
        For analogue ZPK filters, the units in which the zeros and poles are
        specified. Either ``'Hz'`` or ``'rad/s'`` (default).

    normalize_gain : `bool`, optional
        Whether to normalize the gain when converting from Hz to rad/s.

        - `False` (default):
          Multiply zeros/poles by -2π but leave gain unchanged.
          This matches the LIGO GDS **'f' plane** convention
          (``plane='f'`` in ``s2z()``).

        - `True`:
          Normalize gain to preserve frequency response magnitude.
          Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
          Use this when your filter was designed with the transfer
          function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
          This matches the LIGO GDS **'n' plane** convention
          (``plane='n'`` in ``s2z()``).

        Only used for analogue filters in Hz (``analog=True, unit="Hz"``).

    prewarp : `bool`, optional
        If `True`, apply prewarping before bilinear transform (default).
        If `False`, skip prewarping (not recommended).
        Ignored when ``analog=False``.

    output : `str`, optional
        Desired output filter form:

        - ``'zpk'``: zero-pole-gain tuple (default)
        - ``'ba'``: numerator-denominator tuple
        - ``'sos'``: second-order sections array with separated gain

    Returns
    -------
    filt_out : `tuple`
        Filter coefficients in the requested form:

        - ``output='ba'``: 2-tuple of (b, a) arrays
        - ``output='zpk'``: 3-tuple of (zeros, poles, gain)
        - ``output='sos'``: 2-tuple of (sos_array, gain)

        For ``output='sos'``, the returned SOS array has unit gain in all
        sections, and the overall gain is returned separately. This must be
        applied to the filter output: ``filtered_data * gain``.

    Raises
    ------
    ValueError
        If ``analog=True`` but ``sample_rate`` is not provided.

    See Also
    --------
    prepare_analog_filter
        Prepare analog filter without digital conversion.

    scipy.signal.sosfilt
        Apply SOS filter to data.

    Notes
    -----
    **Prewarping**

    Unlike `scipy.signal.bilinear` and `scipy.signal.bilinear_zpk` which
    do NOT perform prewarping, this function applies prewarping by default.
    Prewarping ensures the digital filter's frequency response matches the
    analogue design at the pole/zero frequencies.

    Examples
    --------
    Prepare a digital filter from analog specification:

    >>> from gwpy.signal.filter_design import prepare_digital_filter
    >>> zpk = ([100], [10], 1.0)  # 100 Hz zero, 10 Hz pole, gain 1.0
    >>> sos = prepare_digital_filter(
    ...     zpk,
    ...     analog=True,
    ...     sample_rate=4096,
    ...     unit='Hz',
    ...     output='sos',
    ... )

    Apply to data:

    >>> from scipy import signal
    >>> filtered = signal.sosfilt(sos, data)
    """
    # Parse filter to standardize format
    form, filter_tuple = parse_filter(filt)

    # Convert sample_rate to Hz
    sample_rate = _as_float(sample_rate, unit="Hz")

    # Step 1-2: Prepare analog filter (parse + unit conversion)
    if analog:
        if form == "zpk" and unit == "Hz":
            filter_tuple = cast("ZpkType", filter_tuple)
            filter_tuple = _convert_zpk_units(
                filter_tuple,
                unit=unit,
                normalize_gain=normalize_gain,
            )

        # Convert to ZPK for prewarping and bilinear transform
        zeros, poles, gain = _transform(filter_tuple, form, "zpk")

        # Manual prewarping (GDS default: enabled)
        if prewarp:
            zeros, poles, gain = _prewarp_zpk(zeros, poles, gain, sample_rate)

        # Apply bilinear transform
        filter_tuple = signal.bilinear_zpk(zeros, poles, gain, fs=sample_rate)
        form = "zpk"

    # For ba/zpk output, transform as needed
    return _transform(filter_tuple, form, output)


# -- FIR from transfer function -------

def _truncate_transfer(
    transfer: TArray,
    nsamples: int = 5,
    ncorner: int | None = None,
) -> TArray:
    """Smoothly zero the edges of a frequency domain transfer function.

    Parameters
    ----------
    transfer : `numpy.ndarray`
        Transfer function to start from, must have at least ten samples.

    nsamples : `int`, optional
        Number of samples to taper on each side.

    ncorner : `int`, optional
        Number of extra samples to zero off at low frequency.

    Returns
    -------
    out : `numpy.ndarray`
        The smoothly-truncated transfer function.

    Notes
    -----
    By default, the input transfer function will have five samples tapered
    off at the left and right boundaries. If `ncorner` is not `None`, then
    `ncorner` extra samples will be zeroed on the left as a hard highpass
    filter.

    See Also
    --------
    scipy.signal.windows.tukey
    """
    nsamp = transfer.size
    ncorner = ncorner if ncorner else 0
    out = transfer.copy()
    out[0:ncorner] = 0
    # Apply tukey window to taper edges (5 samples on each side)
    ntaper = nsamp - ncorner
    if ntaper > 0:
        window_full = tukey(2 * nsamples, alpha=1.0)
        out[ncorner:ncorner+nsamples] *= window_full[:nsamples]
        out[nsamp-nsamples:nsamp] *= window_full[-nsamples:]
    return out


def _truncate_impulse(
    impulse: TArray,
    ntaps: int,
    window: WindowLike = "hann",
) -> TArray:
    """Smoothly truncate a time domain impulse response.

    Parameters
    ----------
    impulse : `numpy.ndarray`
        The impulse response to start from.

    ntaps : `int`
        Number of taps in the final filter.

    window : `str`, `float`, `tuple`, `numpy.ndarray`
        Window to truncate with, see `scipy.signal.get_window`
        for details on acceptable formats.

    Returns
    -------
    out : `numpy.ndarray`
        The smoothly truncated impulse response.
    """
    out = impulse.copy()
    trunc_start = int(ntaps / 2)
    trunc_stop = out.size - trunc_start
    window = get_window(window, ntaps)
    out[0:trunc_start] *= window[trunc_start:ntaps]
    out[trunc_stop:out.size] *= window[0:trunc_start]
    out[trunc_start:trunc_stop] = 0
    return out


def fir_from_transfer(
    transfer: Array1D,
    ntaps: int,
    window: WindowLike = "hann",
    ncorner: int | None = None,
) -> NDArray:
    """Design a Type II FIR filter given an arbitrary transfer function.

    Parameters
    ----------
    transfer : `numpy.ndarray`
        Transfer function to start from, must have at least ten samples.

    ntaps : `int`
        Number of taps in the final filter, must be an even number.

    window : `str`, `numpy.ndarray`, optional
        Window function to truncate with, see `scipy.signal.get_window`
        for details on acceptable formats.

    ncorner : `int`, optional
        Number of extra samples to zero off at low frequency.

    Returns
    -------
    out : `numpy.ndarray`
        A time domain FIR filter of length `ntaps`.

    Notes
    -----
    The final FIR filter will use `~numpy.fft.rfft` FFT normalisation.

    If `ncorner` is not `None`, then `ncorner` extra samples will be zeroed
    on the left as a hard highpass filter.

    See Also
    --------
    scipy.signal.remez
        An alternative FIR filter design using the Remez exchange algorithm.
    """
    # truncate and highpass the transfer function
    transfer = _truncate_transfer(transfer, ncorner=ncorner)
    # compute and truncate the impulse response
    impulse = npfft.irfft(transfer)
    impulse = _truncate_impulse(
        impulse,
        ntaps=ntaps,
        window=window,
    )
    # wrap around and normalise to construct the filter
    return numpy.roll(impulse, int(ntaps / 2 - 1))[0:ntaps]


# -- Filter design -------------------

@overload
def _design_iir(
    wp: ArrayLike,
    ws: ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    *,
    analog: bool,
    ftype: FilterTypeName,
    output: Literal["zpk"],
) -> ZpkType: ...

@overload
def _design_iir(
    wp: ArrayLike,
    ws: ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    *,
    analog: bool,
    ftype: FilterTypeName,
    output: Literal["ba"],
) -> BAType: ...

@overload
def _design_iir(
    wp: ArrayLike,
    ws: ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    *,
    analog: bool,
    ftype: FilterTypeName,
    output: Literal["sos"],
) -> SosType: ...

def _design_iir(
    wp: ArrayLike,
    ws: ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    *,
    analog: bool = False,
    ftype: FilterTypeName = "cheby1",
    output: Literal["zpk", "ba", "sos"] = "zpk",
) -> IirFilterType:
    """Design an IIR filter using `scipy.signal.iirdesign`."""
    nyq = sample_rate / 2.
    wp = numpy.atleast_1d(wp)
    ws = numpy.atleast_1d(ws)
    if analog:  # convert Hz to rad/s
        wp *= TWO_PI
        ws *= TWO_PI
    else:  # convert Hz to half-cycles / sample
        wp /= nyq
        ws /= nyq
    z, p, k = signal.iirdesign(
        wp,
        ws,
        gpass,
        gstop,
        analog=analog,
        ftype=ftype,
        output="zpk",
    )
    if analog:  # convert back to Hz
        z /= -TWO_PI
        p /= -TWO_PI
        k *= TWO_PI ** z.size / -TWO_PI ** p.size
    if output == "zpk":
        return z, p, k
    if output == "ba":
        return signal.zpk2tf(z, p, k)
    if output == "sos":
        return signal.zpk2sos(z, p, k)
    msg = f"'{output}' is not a valid output form"
    raise ValueError(msg)


def _design_fir(
    wp: float | ArrayLike,
    ws: float | ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    window: str = "hamming",
    **kwargs,
) -> TapsType:
    """Design an FIR filter using `scipy.signal.firwin`.

    This is just an internal convenience function to calculate the number of
    taps based on the pass and stop band frequencies, and to set sensible
    defaults for a few other keyword arguments.

    See Also
    --------
    scipy.signal.firwin
        for details of how the FIR filters are actually generated
    """
    # format arguments
    wp = numpy.atleast_1d(wp)
    ws = numpy.atleast_1d(ws)
    tw = wp[0] - ws[0]

    # calculate the number of taps
    nt = num_taps(sample_rate, tw, gpass, gstop)

    # set default kw based on the filter shape
    if wp[0] > ws[0]:  # highpass
        kwargs.setdefault("pass_zero", False)
    if ws.shape == (1,):  # simple list of taps
        kwargs.setdefault("width", ws.item() - wp.item())

    kwargs.setdefault("fs", sample_rate)
    return signal.firwin(nt, wp, window=window, **kwargs)


@overload
def _design(
    type_: Literal["iir"],
    wp: float | ArrayLike,
    ws: float | ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    output: Literal["zpk"],
    **kwargs,
) -> ZpkType: ...

@overload
def _design(
    type_: Literal["iir"],
    wp: float | ArrayLike,
    ws: float | ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    output: Literal["ba"],
    **kwargs,
) -> BAType: ...

@overload
def _design(
    type_: Literal["iir"],
    wp: float | ArrayLike,
    ws: float | ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    output: Literal["sos"],
    **kwargs,
) -> SosType: ...

@overload
def _design(
    type_: Literal["fir"],
    wp: float | ArrayLike,
    ws: float | ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    **kwargs,
) -> TapsType: ...

def _design(
    type_: str,
    wp: float | ArrayLike,
    ws: float | ArrayLike,
    sample_rate: float,
    gpass: float,
    gstop: float,
    output: Literal["zpk", "ba", "sos"] = "zpk",
    **kwargs,
) -> FilterType:
    """Design an IIR or FIR filter."""
    designer: Callable
    if type_ == "iir":
        kwargs["output"] = output
        designer = _design_iir
    else:
        designer = _design_fir
    return designer(wp, ws, sample_rate, gpass, gstop, **kwargs)


# -- user methods --------------------

@overload
def lowpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: QuantityLike | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["zpk"] = "zpk",
    **kwargs,
) -> ZpkType: ...

@overload
def lowpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: QuantityLike | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["ba"] = "ba",
    **kwargs,
) -> BAType: ...

@overload
def lowpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: QuantityLike | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["sos"] = "sos",
    **kwargs,
) -> SosType: ...

@overload
def lowpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: QuantityLike | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["fir"] = "fir",
    **kwargs,
) -> TapsType: ...

def lowpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: QuantityLike | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir", "fir"] = "iir",  # noqa: A002
    output: Literal["zpk", "ba", "sos"] = "zpk",
    **kwargs,
) -> FilterType:
    """Design a low-pass filter for the given cutoff frequency.

    Parameters
    ----------
    frequency : `float`
        Corner frequency of low-pass filter (Hertz).

    sample_rate : `float`
        Sampling rate of target data (Hertz).

    fstop : `float`, optional
        Edge-frequency of stop-band (Hertz).

    gpass : `float`, optional, default: 2
        The maximum loss in the passband (dB).

    gstop : `float`, optional, default: 30
        The minimum attenuation in the stopband (dB).

    type : `str`, optional, default: ``'iir'``
        The filter type, either ``'iir'`` or ``'fir'``.

    output : `str`, optional, default: ``'zpk'``
        The output format for an IIR filter,
        either ``'zpk'``, ``'ba'``, or ``'sos'``.

    kwargs
        Other keyword arguments are passed directly to
        :func:`~scipy.signal.iirdesign` or :func:`~scipy.signal.firwin`.

    Returns
    -------
    filter
        The formatted filter.
        The output format for an IIR filter depends on the input arguments,
        default is a tuple of ``(zeros, poles, gain)``.

    Notes
    -----
    By default a digital filter is returned, meaning the zeros and poles
    are given in the Z-domain in units of radians/sample.

    Examples
    --------
    To create a low-pass filter at 1000 Hz for 4096 Hz-sampled data:

    >>> from gwpy.signal.filter_design import lowpass
    >>> lp = lowpass(1000, 4096)

    To view the filter, you can use the `~gwpy.plot.BodePlot`:

    >>> from gwpy.plot import BodePlot
    >>> plot = BodePlot(lp, sample_rate=4096)
    >>> plot.show()
    """
    sample_rate = _as_float(sample_rate)
    frequency = _as_float(frequency)
    if fstop is None:
        fstop = min(frequency * 1.5, sample_rate / 2.)
    fstop = _as_float(fstop)
    return _design(
        type,
        frequency,
        fstop,
        sample_rate,
        gpass,
        gstop,
        output=output,
        **kwargs,
    )


@overload
def highpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: float | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["zpk"] = "zpk",
    **kwargs,
) -> ZpkType: ...

@overload
def highpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: float | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["ba"] = "ba",
    **kwargs,
) -> BAType: ...

@overload
def highpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: float | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["sos"] = "sos",
    **kwargs,
) -> SosType: ...

@overload
def highpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: float | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["fir"] = "fir",
    **kwargs,
) -> TapsType: ...

def highpass(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    fstop: float | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir", "fir"] = "iir",  # noqa: A002
    output: Literal["zpk", "ba", "sos"] = "zpk",
    **kwargs,
) -> FilterType:
    """Design a high-pass filter for the given cutoff frequency.

    Parameters
    ----------
    frequency : `float`, `~astropy.units.Quantity`
        Corner frequency of high-pass filter.

    sample_rate : `float`, `~astropy.units.Quantity`
        Sampling rate of target data.

    fstop : `float`
        Edge-frequency of stop-band.

    gpass : `float`
        The maximum loss in the passband (dB)

    gstop : `float`
        The minimum attenuation in the stopband (dB).

    type : `str`
        The filter type, either ``'iir'`` or ``'fir'``.

    output : `str`, optional, default: ``'zpk'``
        The output format for an IIR filter,
        either ``'zpk'``, ``'ba'``, or ``'sos'``.

    kwargs
        Other keyword arguments are passed directly to
        :func:`~scipy.signal.iirdesign` or :func:`~scipy.signal.firwin`.

    Returns
    -------
    filter
        The formatted filter.
        The output format for an IIR filter depends on the input arguments,
        default is a tuple of ``(zeros, poles, gain)``.

    Notes
    -----
    By default a digital filter is returned, meaning the zeros and poles
    are given in the Z-domain in units of radians/sample.

    Examples
    --------
    To create a high-pass filter at 100 Hz for 4096 Hz-sampled data:

    >>> from gwpy.signal.filter_design import highpass
    >>> hp = highpass(100, 4096)

    To view the filter, you can use the `~gwpy.plot.BodePlot`:

    >>> from gwpy.plot import BodePlot
    >>> plot = BodePlot(hp, sample_rate=4096)
    >>> plot.show()
    """
    sample_rate = _as_float(sample_rate)
    frequency = _as_float(frequency)
    if fstop is None:
        fstop = frequency * 2 / 3.
    return _design(
        type,
        frequency,
        fstop,
        sample_rate,
        gpass,
        gstop,
        output=output,
        **kwargs,
    )


@overload
def bandpass(
    flow: QuantityLike,
    fhigh: QuantityLike,
    sample_rate: QuantityLike,
    fstop: tuple[QuantityLike, QuantityLike] | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["zpk"] = "zpk",
    **kwargs,
) -> ZpkType: ...

@overload
def bandpass(
    flow: QuantityLike,
    fhigh: QuantityLike,
    sample_rate: QuantityLike,
    fstop: tuple[QuantityLike, QuantityLike] | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["ba"] = "ba",
    **kwargs,
) -> BAType: ...

@overload
def bandpass(
    flow: QuantityLike,
    fhigh: QuantityLike,
    sample_rate: QuantityLike,
    fstop: tuple[QuantityLike, QuantityLike] | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir"] = "iir",
    output: Literal["sos"] = "sos",
    **kwargs,
) -> SosType: ...

@overload
def bandpass(
    flow: QuantityLike,
    fhigh: QuantityLike,
    sample_rate: QuantityLike,
    fstop: tuple[QuantityLike, QuantityLike] | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["fir"] = "fir",
    **kwargs,
) -> TapsType: ...

def bandpass(
    flow: QuantityLike,
    fhigh: QuantityLike,
    sample_rate: QuantityLike,
    fstop: tuple[QuantityLike, QuantityLike] | None = None,
    gpass: float = 2,
    gstop: float = 30,
    type: Literal["iir", "fir"] = "iir",  # noqa: A002
    output: Literal["zpk", "ba", "sos"] = "zpk",
    **kwargs,
) -> FilterType:
    """Design a band-pass filter for the given cutoff frequencies.

    Parameters
    ----------
    flow : `float`, `~astropy.units.Quantity`
        Lower corner frequency of pass band.

    fhigh : `float`, `~astropy.units.Quantity`
        Upper corner frequency of pass band.

    sample_rate : `float`, `~astropy.units.Quantity`
        Sampling rate of target data.

    fstop : `tuple` of `float`
        `(low, high)` edge-frequencies of stop band.

    gpass : `float`
        The maximum loss in the passband (dB).

    gstop : `float`
        The minimum attenuation in the stopband (dB).

    type : `str`
        The filter type, either ``'iir'`` or ``'fir'``.

    output : `str`, optional, default: ``'zpk'``
        The output format for an IIR filter,
        either ``'zpk'``, ``'ba'``, or ``'sos'``.

    kwargs
        Other keyword arguments are passed directly to
        :func:`~scipy.signal.iirdesign` or :func:`~scipy.signal.firwin`.

    Returns
    -------
    filter
        The formatted filter.
        The output format for an IIR filter depends on the input arguments,
        default is a tuple of `(zeros, poles, gain)`.

    Notes
    -----
    By default a digital filter is returned, meaning the zeros and poles
    are given in the Z-domain in units of radians/sample.

    Examples
    --------
    To create a band-pass filter for 100-1000 Hz for 4096 Hz-sampled data:

    >>> from gwpy.signal.filter_design import bandpass
    >>> bp = bandpass(100, 1000, 4096)

    To view the filter, you can use the `~gwpy.plot.BodePlot`:

    >>> from gwpy.plot import BodePlot
    >>> plot = BodePlot(bp, sample_rate=4096)
    >>> plot.show()
    """
    sample_rate = _as_float(sample_rate)
    flow = _as_float(flow)
    fhigh = _as_float(fhigh)
    if fstop is None:
        fstop = (
            flow * 2 / 3.,
            min(fhigh * 1.5, sample_rate / 2.),
        )
    fstop = (_as_float(fstop[0]), _as_float(fstop[1]))
    if type == "fir":
        kwargs.setdefault("pass_zero", False)
    return _design(
        type,
        [flow, fhigh],
        fstop,
        sample_rate,
        gpass,
        gstop,
        output=output,
        **kwargs,
    )


@overload
def notch(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    type: Literal["iir"] = "iir",
    output: Literal["zpk"] = "zpk",
    **kwargs,
) -> ZpkType: ...

@overload
def notch(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    type: Literal["iir"] = "iir",
    output: Literal["ba"] = "ba",
    **kwargs,
) -> BAType: ...

@overload
def notch(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    type: Literal["iir"] = "iir",
    output: Literal["sos"] = "sos",
    **kwargs,
) -> SosType: ...

def notch(
    frequency: QuantityLike,
    sample_rate: QuantityLike,
    type: Literal["iir"] = "iir",  # noqa: A002
    output: Literal["zpk", "ba", "sos"] = "zpk",
    **kwargs,
) -> IirFilterType:
    """Design a ZPK notch filter for the given frequency and sampling rate.

    Parameters
    ----------
    frequency : `float`, `~astropy.units.Quantity`
        Frequency (default in Hertz) at which to apply the notch.

    sample_rate : `float`, `~astropy.units.Quantity`
        Number of samples per second for `TimeSeries` to which this notch
        filter will be applied.

    type : `str`, optional, default: 'iir'
        Type of filter to apply, currently only 'iir' is supported.

    output : `str`, optional, default: 'zpk'
        Output format for notch.

    kwargs
        Other keyword arguments to pass to `scipy.signal.iirdesign`.

    Returns
    -------
    filter
        The formatted filter.
        The output format for an IIR filter depends on the input arguments,
        default is a tuple of ``(zeros, poles, gain)``.

    See Also
    --------
    scipy.signal.iirdesign
        For details on the IIR filter design method and the output formats.

    Notes
    -----
    By default a digital filter is returned, meaning the zeros and poles
    are given in the Z-domain in units of radians/sample.

    Examples
    --------
    To create a low-pass filter at 1000 Hz for 4096 Hz-sampled data:

    >>> from gwpy.signal.filter_design import notch
    >>> n = notch(100, 4096)

    To view the filter, you can use the `~gwpy.plot.BodePlot`:

    >>> from gwpy.plot import BodePlot
    >>> plot = BodePlot(n, sample_rate=4096)
    >>> plot.show()
    """
    frequency = _as_float(frequency)
    sample_rate = _as_float(sample_rate)
    df = 1.0
    df2 = 0.1
    low1 = (frequency - df)
    high1 = (frequency + df)
    low2 = (frequency - df2)
    high2 = (frequency + df2)
    if type == "iir":
        kwargs.setdefault("gpass", 1)
        kwargs.setdefault("gstop", 10)
        kwargs.setdefault("ftype", "ellip")
        return _design_iir(
            [low1, high1],
            [low2, high2],
            sample_rate,
            output=output,
            **kwargs,
        )
    msg = f"Generating {type} notch filters has not been implemented yet"
    raise NotImplementedError(msg)


# -- Frequency response --------------

def frequency_response(
    filt: FilterCompatible,
    frequencies: NDArray | int | None,
    *,
    analog: bool = False,
    sample_rate: QuantityLike = 1.0,
    unit: str | UnitBase = "rad/s",
    normalize_gain: bool = False,
) -> tuple[numpy.ndarray[tuple[int]], numpy.ndarray[tuple[int]]]:
    """Compute the frequency response of a filter at given frequencies.

    Parameters
    ----------
    filt : `FilterCompatible`
        Filter definition in one of the supported formats.

    frequencies : `numpy.ndarray`
        Frequencies (in Hz) at which to compute the response.

    analog : `bool`, optional
        Whether the filter is analogue (`True`) or digital (`False`).

    sample_rate : `astropy.units.Quantity`-like, optional
        Sample rate of the digital data (only used if `analog` is `False`).

    unit : `str`, optional
        For analogue ZPK filters, the units in which the zeros and poles are
        specified. Either ``'Hz'`` or ``'rad/s'`` (default).

    normalize_gain : `bool`, optional
        Whether to normalize the gain when converting from Hz to rad/s.

        - `False` (default):
          Multiply zeros/poles by -2π but leave gain unchanged.
          This matches the LIGO GDS **'f' plane** convention
          (``plane='f'`` in ``s2z()``).

        - `True`:
          Normalize gain to preserve frequency response magnitude.
          Gain is scaled by :math:`|∏p_i/∏z_i| · (2π)^{(n_p - n_z)}`.
          Use this when your filter was designed with the transfer
          function :math:`H(f) = k·∏(f-z_i)/∏(f-p_i)` in Hz.
          This matches the LIGO GDS **'n' plane** convention
          (``plane='n'`` in ``s2z()``).

        Only used for analogue filters in Hz (``analog=True, unit="Hz"``).

    Returns
    -------
    frequencies : `numpy.ndarray`
        Frequencies at which the response was computed.
        If ``analog=True, unit='rad/s'``, these are in rad/s, otherwise
        they are in Hz.
    response : `numpy.ndarray`
        Frequency response of the filter at the given frequencies.

    See Also
    --------
    scipy.signal.freqs
    scipy.signal.freqz
    scipy.signal.freqz_sos
    scipy.signal.freqs_zpk
    scipy.signal.freqz_zpk
    """
    filt_: BAType | ZpkType | SosType
    if analog:
        form, filt_ = prepare_analog_filter(
            filt,
            unit=unit,
            normalize_gain=normalize_gain,
        )
    else:
        form = "zpk"
        filt_ = prepare_digital_filter(
            filt,
            sample_rate=sample_rate,
            unit=unit,
            normalize_gain=normalize_gain,
            output=form,
        )

    # Use angular frequencies
    if frequencies is not None and not isinstance(frequencies, int):
        frequencies *= TWO_PI

    # Compute the filter response
    if analog and form == "sos":
        msg = "analog SOS frequency response not supported"
        raise ValueError(msg)

    # Digital SOS
    if form == "sos":
        wfreq, fresp = signal.freqz_sos(filt_, worN=frequencies)
    # Analogue ZPK
    elif analog and form == "zpk":
        filt_ = cast("ZpkType", filt_)
        wfreq, fresp = signal.freqs_zpk(*filt_, worN=frequencies)
    # Digital ZPK
    elif form == "zpk":
        filt_ = cast("ZpkType", filt_)
        wfreq, fresp = signal.freqz_zpk(*filt_, worN=frequencies)
    # Analogue BA
    elif analog:
        filt_ = cast("BAType", filt_)
        wfreq, fresp = signal.freqs(*filt_, worN=frequencies)
    # Digital BA
    else:
        filt_ = cast("BAType", filt_)
        wfreq, fresp = signal.freqz(*filt_, worN=frequencies)

    # Convert from digital angular frequency to Hz
    if not analog:
        wfreq *= _as_float(sample_rate, unit="Hz") / TWO_PI
    elif unit == "Hz":
        wfreq /= TWO_PI


    return (
        wfreq,
        fresp,
    )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for auto-discovery of GW data files.

Automatic discovery of file paths for both LIGO and Virgo index solutions
(``gwdatafind`` or ``ffl``, respectvely) is supported.

The functions in this module are highly reliant on having local access to
files (either directly, or via NFS/fuse).

Data discovery using the DataFind service requires the `gwdatafind` Python
package (a dependency of ``gwpy``), and either the ``GW_DATAFIND_SERVER``
(or legacy ``LIGO_DATAFIND_SERVER``) environment variable to be set,
or the ``host`` keyword must be passed to :func:`find_urls` and friends.

Data discovery using the Virgo FFL system requires the ``FFLPATH`` environment
variable to point to the directory containing FFL files, **or** the
``VIRGODATA`` environment variable to point to a directory containing an
``ffl`` subdirectory, which contains FFL files.
"""

from __future__ import annotations

import logging
import os
import re
from collections import defaultdict
from contextlib import nullcontext
from functools import (
    partial,
    wraps,
)
from math import ceil
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    cast,
    overload,
)
from unittest import mock

import gwdatafind
from igwn_segments import segment as LigoSegment  # noqa: N812
from urllib3.util import parse_url

from ..detector import Channel
from ..time import to_gps
from . import ffldatafind
from .cache import cache_segments
from .gwf import (
    iter_channel_names,
    num_channels,
)
from .remote import (
    NETWORK_ERROR,
    download_file,
    is_remote,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
        Mapping,
    )
    from types import ModuleType
    from typing import (
        Literal,
        ParamSpec,
        SupportsFloat,
        TypeVar,
    )

    from ..segments import Segment
    from ..time import SupportsToGps

    P = ParamSpec("P")
    T = TypeVar("T")

    ChannelLike = TypeVar("ChannelLike", bound=str | Channel)

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "find_best_frametype",
    "find_frametype",
    "find_latest",
    "find_types",
    "find_urls",
    "on_tape",
]

logger = logging.getLogger(__name__)

SINGLE_IFO_OBSERVATORY = re.compile("^[A-Z][0-9]$")

GPS_END_S5 = 875232014

# special-case frame types
LIGO_SECOND_TREND_TYPE = re.compile(r"\A(.*_)?T\Z")  # T or *_T
LIGO_MINUTE_TREND_TYPE = re.compile(r"\A(.*_)?M\Z")  # M or *_M
VIRGO_SECOND_TREND_TYPE = re.compile(r"\A(.*_)?[Tt]rend\Z")  # trend or *_trend
NOT_GRB_TYPE = re.compile(r"^(?!.*_GRB\d{6}([A-Z])?$)")  # *_GRBYYMMDD{A}
HIGH_PRIORITY_TYPE = re.compile("({})".format("|".join((  # noqa: FLY002
    r"\A[A-Z]\d_GWOSC_O([0-9]+)([a-z])?_[0-9]+KHZ_R[0-9]+",  # X1_GWOSC_OX_NKHZ_RX
    r"\A[A-Z]\d_HOFT_C\d\d(_T\d{7}_v\d)?\Z",  # X1_HOFT_CXY
    r"\AV1Online\Z",
    r"\AHoftOnline\Z",
    r"\AV1O[0-9]+([A-Z]+)?Repro[0-9]+[A-Z]+\Z",  # V1OXReproXY
))))
LOW_PRIORITY_TYPE = re.compile("({})".format("|".join((  # noqa: FLY002
    r"_GRB\d{6}([A-Z])?\Z",  # endswith _GRBYYMMDD{A}
    r"_bck\Z",  # endswith _bck
    r"\AT1200307_V4_EARLY_RECOLORED_V2\Z",  # annoying recoloured HOFT type
))))


# -- utilities -----------------------

def _type_priority(
    ftype: str,
    trend: str | None = None,
) -> tuple[int, int]:
    """Prioritise the given GWF type based on its name or trend status.

    This is essentially an ad-hoc ordering function based on internal knowledge
    of how LIGO does GWF type naming.
    """
    # if looking for a trend channel, prioritise the matching type
    for trendname, trend_regex in [
        ("m-trend", LIGO_MINUTE_TREND_TYPE),
        ("s-trend", LIGO_SECOND_TREND_TYPE),
        ("s-trend", VIRGO_SECOND_TREND_TYPE),
    ]:
        if trend == trendname and trend_regex.match(ftype):
            return 0, len(ftype)

    # otherwise rank this type according to priority
    for reg, prio in {
        HIGH_PRIORITY_TYPE: 1,
        re.compile(r"[A-Z]\d_C"): 6,
        LOW_PRIORITY_TYPE: 10,
        LIGO_MINUTE_TREND_TYPE: 10,
        LIGO_SECOND_TREND_TYPE: 10,
        VIRGO_SECOND_TREND_TYPE: 10,
    }.items():
        if reg.search(ftype):
            return prio, len(ftype)

    return 5, len(ftype)


def on_tape(*files: str) -> bool:
    """Determine whether any of the given files are on tape.

    Parameters
    ----------
    *files : `str`
        one or more paths to GWF files

    Returns
    -------
    ontape : `bool`
        `True` if any of the files are determined to be on tape,
        otherwise `False`.

    Notes
    -----
    A returned value of `False` does not necessarily mean that none of the
    paths are on tape, just that this function couldn't say that they
    definitely *are* on tape.
    """
    for path in files:
        if is_remote(path):  # we can't inspect remote files
            return False
        url = parse_url(path)
        if not url.path:
            return False
        try:
            stat = Path(url.path).stat()
        except FileNotFoundError:
            return False
        try:
            return stat.st_blocks == 0
        except AttributeError:  # windows doesn't have st_blocks
            return False
    return False


def _gwdatafind_module(**datafind_kw) -> ModuleType:
    """Return the appropriate GWDataFind-like API based on the environment.

    This allows switching to the hacky `gwpy.io.ffldatafind` replacement
    module to enable a GWDataFind-like interface for direct FFL data
    discovery at Virgo.
    """
    # GWDataFind
    if (
        os.getenv("GWDATAFIND_SERVER")
        or os.getenv("LIGO_DATAFIND_SERVER")
        or datafind_kw.get("host")
    ):
        return gwdatafind

    # FFL
    try:
        ffldatafind._get_ffl_basedir()  # noqa: SLF001
    except KeyError:  # failed to discover FFL directories
        msg = "unknown datafind configuration, cannot discover data"
        raise RuntimeError(msg) from None
    return ffldatafind


def _select_gwdatafind_mod(func: Callable[P, T]) -> Callable[P, T]:
    """Decorate a function to see the right ``gwdatafind`` API.

    This exists only to allow on-the-fly replacing of the actual `gwdatafind`
    with :mod:`gwpy.io.ffldatafind` if it looks like we are trying to find
    data from FFL files.
    """
    @wraps(func)
    def wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        # replace the 'gwdatafind' module in the function namespace
        # with the API we need for this call
        with mock.patch.dict(func.__globals__):
            func.__globals__["gwdatafind"] = _gwdatafind_module(**kwargs)
            return func(*args, **kwargs)

    return wrapped


def _parse_ifos_and_trends(
    chans: Iterable[str | Channel],
) -> set[tuple[str, str | None]]:
    """Parse ``(ifo, trend)`` pairs from this list of channels."""
    found = set()
    for name in chans:
        chan = Channel(name)
        if (ifo := chan.ifo) is None:
            msg = (
                "Cannot parse interferometer prefix from channel name "
                f"'{chan}', cannot proceed with find()"
            )
            raise ValueError(msg)
        found.add((ifo[0], chan.type))
    return found


def _find_gaps(
    ifo: str,
    frametype: str,
    segment: Segment | None,
    **kwargs,
) -> float:
    """Discover gaps in a datafind/ffl archive for the given ifo/type.

    Returns
    -------
    gaps : `float`
        The cumulative size of all gaps in the relevant archive.
    """
    if segment is None:
        return 0
    cache = find_urls(
        ifo,
        frametype,
        *segment,
        **kwargs,
    )
    csegs = cache_segments(cache)
    return max(0, abs(segment) - abs(csegs))


def _error_missing_channels(
    required: Iterable[ChannelLike],
    found: Iterable[ChannelLike],
    gpstime: SupportsFloat | None,
    *,
    allow_tape: bool,
) -> None:
    """Raise an exception if required channels are not found."""
    missing = set(map(str, required)) - set(map(str, found))

    if not missing:  # success
        return

    # failure
    msg = "Cannot locate the following channel(s) in any known frametype"
    if gpstime:
        msg += f" at GPS={gpstime}"
    msg += ":\n    " + "\n    ".join(missing)
    if not allow_tape:
        msg += (
            "\n[files on tape have not been checked, use "
            "allow_tape=True for a complete search]"
        )
    raise ValueError(msg)


def _rank_types(
    match: Mapping[ChannelLike, list[tuple[str, str, float]]],
) -> None:
    """Rank and sort the matched frametypes according to some criteria.

    Parameters
    ----------
    match : `dict` of ``channel: (type, gwf, gapsize)``
        The match dict to sort.
    """
    paths = {typetuple[1] for key in match for typetuple in match[key]}
    rank = {path: (on_tape(path), num_channels(path)) for path in paths}
    # deprioritise types on tape and those with lots of channels
    for key in match:
        match[key].sort(key=lambda x: (-x[2],) + rank[x[1]])


# -- user methods --------------------

# single channel, return_all=False
@overload
def find_frametype(
    channel: str | Channel,
    gpstime: SupportsToGps | None = None,
    *,
    frametype_match: str | re.Pattern | None = None,
    host: str | None = None,
    urltype: str = "file",
    ext: str = "gwf",
    return_all: Literal[False] = False,
    allow_tape: bool = False,
    on_gaps: Literal["error", "ignore", "warn"] = "error",
    **gwdatafind_kw,
) -> str: ...

# single channel, return_all=True
@overload
def find_frametype(
    channel: str | Channel,
    gpstime: SupportsToGps | None = None,
    *,
    frametype_match: str | re.Pattern | None = None,
    host: str | None = None,
    urltype: str = "file",
    ext: str = "gwf",
    return_all: Literal[True] = True,
    allow_tape: bool = False,
    on_gaps: Literal["error", "ignore", "warn"] = "error",
    **gwdatafind_kw,
) -> list[str]: ...

# single channel, return_all not given
@overload
def find_frametype(
    channel: str | Channel,
    gpstime: SupportsToGps | None = None,
    *,
    frametype_match: str | re.Pattern | None = None,
    host: str | None = None,
    urltype: str = "file",
    ext: str = "gwf",
    allow_tape: bool = False,
    on_gaps: Literal["error", "ignore", "warn"] = "error",
    **gwdatafind_kw,
) -> str: ...

# multiple channels, return_all=False
@overload
def find_frametype(
    channel: Iterable[ChannelLike],
    gpstime: SupportsToGps | None = None,
    *,
    frametype_match: str | re.Pattern | None = None,
    host: str | None = None,
    urltype: str = "file",
    ext: str = "gwf",
    return_all: Literal[False] = False,
    allow_tape: bool = False,
    on_gaps: Literal["error", "ignore", "warn"] = "error",
    **gwdatafind_kw,
) -> dict[ChannelLike, str]: ...

# multiple channels, return_all=True
@overload
def find_frametype(
    channel: Iterable[ChannelLike],
    gpstime: SupportsToGps | None = None,
    *,
    frametype_match: str | re.Pattern | None = None,
    host: str | None = None,
    urltype: str = "file",
    ext: str = "gwf",
    return_all: Literal[True] = True,
    allow_tape: bool = False,
    on_gaps: Literal["error", "ignore", "warn"] = "error",
    **gwdatafind_kw,
) -> dict[ChannelLike, list[str]]: ...

# multiple channels, return_all not given
@overload
def find_frametype(
    channel: Iterable[ChannelLike],
    gpstime: SupportsToGps | None = None,
    *,
    frametype_match: str | re.Pattern | None = None,
    host: str | None = None,
    urltype: str = "file",
    ext: str = "gwf",
    allow_tape: bool = False,
    on_gaps: Literal["error", "ignore", "warn"] = "error",
    **gwdatafind_kw,
) -> dict[ChannelLike, list[str]]: ...

def find_frametype(
    channel: str | Channel | Iterable[ChannelLike],
    gpstime: SupportsToGps | None = None,
    *,
    frametype_match: str | re.Pattern | None = None,
    urltype: str = "file",
    ext: str = "gwf",
    return_all: bool = False,
    allow_tape: bool = False,
    cache: bool | None = None,
    on_gaps: Literal["error", "ignore", "warn"] = "error",
    **gwdatafind_kw,
) -> str | list[str] | dict[ChannelLike, str] | dict[ChannelLike, list[str]]:
    """Find the frametype(s) that hold data for a given channel.

    Parameters
    ----------
    channel : `str`, `~gwpy.detector.Channel`
        The channel to find.

    gpstime : `int`, optional
        Target GPS time at which to find correct type.

    frametype_match : `str`, optional
        Regular expression to use for frametype `str` matching.

    host : `str`, optional
        Name of datafind host to use.

    urltype : `str`, optional
        The URL type to use.
        Default is "file" to use paths available on the file system.

    ext : `str`, optional
        The file extension for which to search.
        "gwf" is the only file extension supported, but this may be
        extended in the future.

    return_all : `bool`, optional
        If `True` return all found types;
        if `False` (default) return only the 'best' match.

    allow_tape : `bool`, optional
        If `False` (default) do not test types whose frame files are
        stored on tape (not on spinning disk).

    cache : `bool`, `None`, optional
        Whether to cache the contents of remote URLs.
        Default (`None`) is to check the ``GWPY_CACHE`` environment variable.
        See :ref:`gwpy-env-variables` for details.

    on_gaps : `str`, optional
        Action to take when the requested all or some of the GPS interval
        is not covereed by the dataset, one of:

        - ``'error'``: raise a `RuntimeError` (default)
        - ``'warn'``: emit a warning but return all available URLs
        - ``'ignore'``: return the list of URLs with no warnings

    gwdatafind_kw
        Other keyword arguments are passed to the
        `gwdatafind.find_types`, and `gwdatafind.find_urls`.functions.

    Returns
    -------
    If a single name is given, and `return_all=False` (default):

    frametype : `str`
        name of best match frame type

    If a single name is given, and `return_all=True`:

    types : `list` of `str`
        the list of all matching frame types

    If multiple names are given, the above return types are wrapped into a
    `dict[str, str | list[str]]`.

    Examples
    --------
    >>> from gwpy.io import datafind as io_datafind
    >>> io_datafind.find_frametype('H1:IMC-PWR_IN_OUTPUT', gpstime=1126259462)
    'H1_R'
    >>> io_datafind.find_frametype(
    ...     'H1:IMC-PWR_IN_OUTPUT',
    ...     gpstime=1126259462,
    ...     return_all=True,
    ... )
    ['H1_R', 'H1_C']
    >>> io_datafind.find_frametype(
    ...     ('H1:IMC-PWR_IN_OUTPUT', 'H1:OMC-DCPD_SUM_OUTPUT', 'H1:GDS-CALIB_STRAIN'),
    ...     gpstime=1126259462,
    ...     return_all=True,
    ... )
    {'H1:GDS-CALIB_STRAIN': ['H1_HOFT_C00'],
     'H1:OMC-DCPD_SUM_OUTPUT': ['H1_R', 'H1_C'],
     'H1:IMC-PWR_IN_OUTPUT': ['H1_R', 'H1_C']}
    """
    # this function is now horrendously complicated to support a large
    # number of different use cases, hopefully the comments are sufficient

    # check required file extension
    if ext.lower().lstrip(".") != "gwf":
        msg = "Finding frametypes for channels is only supported for ext='gwf'"
        raise ValueError(msg)
    ext = ext.lower().lstrip(".")

    # check required file extension
    if ext.lower().lstrip(".") != "gwf":
        msg = "Finding frametypes for channels is only supported for ext='gwf'"
        raise ValueError(msg)
    ext = ext.lower().lstrip(".")

    # format channel names as list
    if isinstance(channel, list | tuple):
        channels = channel
    else:
        channels = [channel]

    logger.debug("Finding frametype for channels: %s", ",".join(channels))

    # create set() of GWF channel names, and dict map back to user names
    #    this allows users to use nds-style names in this query, e.g.
    #    'X1:TEST.mean,m-trend', and still get results
    chandict: dict[ChannelLike, str] = {c: str(Channel(c).name) for c in channels}
    names = {val: key for key, val in chandict.items()}

    # format GPS time(s)
    if isinstance(gpstime, tuple):
        gpssegment = LigoSegment(*gpstime)
        gpstime = gpssegment[0]
    else:
        gpssegment = None
    if gpstime is not None:
        gpstime = int(to_gps(gpstime))

    # if use gaps post-S5 GPStime, forcibly skip _GRBYYMMDD frametypes at CIT
    if frametype_match is None and gpstime is not None and gpstime > GPS_END_S5:
        frametype_match = NOT_GRB_TYPE

    # -- go

    match: dict[ChannelLike, list[tuple[str, str, float]]] = defaultdict(list)
    searched = set()

    if sess := gwdatafind_kw.pop("session", None):
        ctx = nullcontext(sess)
    elif _gwdatafind_module(**gwdatafind_kw) is gwdatafind:
        ctx = gwdatafind.Session()
    else:
        ctx = nullcontext()

    with ctx as sess:
        if sess:
            gwdatafind_kw["session"] = sess

        for ifo, trend in _parse_ifos_and_trends(channels):
            logger.debug("Finding types for %s", ifo)

            # find all types (prioritising trends if we need to)
            types = find_types(
                ifo,
                match=frametype_match,
                trend=trend,
                ext=ext,
                **gwdatafind_kw,
            )

            logger.debug("Found %s types", len(types))

            # loop over types testing each in turn
            for ftype in types:

                # if we've already search this type for this IFO,
                # don't do it again
                if (ifo, ftype) in searched:
                    continue

                thismatch = _inspect_ftype(
                    list(names),
                    ifo,
                    ftype,
                    gpstime,
                    gpssegment,
                    on_gaps,
                    allow_tape=allow_tape,
                    urltype=urltype,
                    ext=ext,
                    cache=cache,
                    **gwdatafind_kw,
                )

                if thismatch is None:  # failed to read
                    continue

                for name, info in thismatch.items():
                    n = names[name]
                    match[n].append(info)

                    # if only matching once, don't search other types
                    # for this channel
                    if not return_all:
                        names.pop(name)

                # record this type as having been searched
                searched.add((ifo, ftype))

                if not names:  # if all channels matched, stop
                    break

    # raise exception if one or more channels were not found
    _error_missing_channels(
        names.values(),
        match.keys(),
        gpstime,
        allow_tape=allow_tape or urltype != "file",
    )

    # rank types (and pick best if required)
    _rank_types(match)

    # and format as a dict for each channel
    results: dict[ChannelLike, list[str]] = {
        key: list(next(zip(*match[key], strict=True)))
        for key in match
    }

    # -- handle various return scenarios

    # multiple channels, return_all=True
    if isinstance(channel, list | tuple) and return_all:
        return results

    # multiple channels, return_all=False
    if isinstance(channel, list | tuple):
        return {key: val[0] for key, val in results.items()}

    # single channel, return_all=True
    channel = cast("ChannelLike", channel)
    single = results[channel]
    if return_all:
        return single

    # single channel, return_all=False
    return single[0]


def _inspect_ftype(
    names: list[str],
    ifo: str,
    frametype: str,
    gpstime: int | None,
    gpssegment: Segment | None,
    on_gaps: Literal["error", "ignore", "warn"],
    *,
    allow_tape: bool = False,
    cache: bool | None = None,
    **requests_kw,
) -> dict[str, tuple[str, str, float]] | None:
    """Inspect one dataset (frametype) for matches to the required ``names``.

    This function queries GWDataFind for a single file matching this dataset,
    downloads it, and looks at the list of channel names in the TOC.
    """
    logger.debug("Inspecting %s-%s", ifo, frametype)
    # find instance of this frametype
    try:
        path = find_latest(
            ifo,
            frametype,
            gpstime=gpstime,
            allow_tape=allow_tape,
            **requests_kw,
        )
    except (OSError, RuntimeError, IndexError):  # something went wrong
        logger.debug("Failed to find URL for %s-%s", ifo, frametype)
        return None

    # download the file so we can inspect it
    logger.debug("Using URL '%s'", path)
    try:
        path = download_file(path, cache=cache)
    except NETWORK_ERROR as exc:  # failed to download the file
        logger.debug(
            "Failed to download file for %s-%s: %s",
            ifo,
            frametype,
            str(exc),
        )
        return None

    # check for gaps in the record for this type
    gaps = _find_gaps(
        ifo,
        frametype,
        gpssegment,
        on_gaps=on_gaps,
        **requests_kw,
    )

    # record matches for each channel
    match: dict[str, tuple[str, str, float]] = {}

    # search the TOC for one frame file and match any channels
    found = 0
    nchan = len(names)
    try:
        for n in iter_channel_names(path):
            if n in names:  # frametype includes this channel!
                # count how many channels we have found in this type
                found += 1

                # record the match using the user-given channel name
                match[n] = (frametype, path, gaps)

                if found == nchan:  # all channels have been found
                    break

    except (
        OSError,
        RuntimeError,
    ) as exc:  # failed to open file (probably)
        logger.warning(
            "failed to read channels for %s-%s: %s",
            ifo,
            frametype,
            str(exc),
        )
        return None

    logger.debug(
        "Found %s/%s channels in %s-%s",
        len(match),
        nchan,
        ifo,
        frametype,
    )
    return match


@overload
def find_best_frametype(
    channel: str | Channel,
    start: SupportsToGps,
    end: SupportsToGps,
    *,
    allow_tape: bool = True,
    **kwargs,
) -> str:
    ...


@overload
def find_best_frametype(
    channel: Iterable[ChannelLike],
    start: SupportsToGps,
    end: SupportsToGps,
    *,
    allow_tape: bool = True,
    **kwargs,
) -> dict[ChannelLike, str]:
    ...


def find_best_frametype(
    channel: str | Channel | Iterable[ChannelLike],
    start: SupportsToGps,
    end: SupportsToGps,
    *,
    allow_tape: bool = True,
    **kwargs,
) -> str | dict[ChannelLike, str]:
    """Intelligently select the best frametype from which to read this channel.

    Parameters
    ----------
    channel : `str`, `~gwpy.detector.Channel`
        the channel to be found

    start : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS start time of period of interest,
        any input parseable by `~gwpy.time.to_gps` is fine

    end : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS end time of period of interest,
        any input parseable by `~gwpy.time.to_gps` is fine

    allow_tape : `bool`, optional
        do not test types whose frame files are stored on tape (not on
        spinning disk)

    kwargs
        Other keyword arguments are passed to `find_frametype`.

    Returns
    -------
    frametype : `str`
        the best matching frametype for the ``channel`` in the
        ``[start, end)`` interval

    Raises
    ------
    ValueError
        If no valid frametypes are found.

    See Also
    --------
    find_frametype
        For details of how available frametypes are matched.

    Examples
    --------
    >>> from gwpy.io.datafind import find_best_frametype
    >>> find_best_frametype('L1:GDS-CALIB_STRAIN', 1126259460, 1126259464)
    'L1_HOFT_C00'
    """
    try:
        # try and find the one true frametype for this channel
        return find_frametype(
            channel,
            gpstime=(start, end),
            allow_tape=allow_tape,
            return_all=False,
            on_gaps="error",
            **kwargs,
        )
    except RuntimeError:  # gaps (or something else went wrong)
        # find all frametypes
        ftout = find_frametype(
            channel,
            gpstime=(start, end),
            allow_tape=allow_tape,
            return_all=True,
            on_gaps="ignore",
            **kwargs,
        )
        # and pick the one with the highest rank
        # (as applied by find_frametype)
        try:
            if isinstance(ftout, dict):
                return {key: ftout[key][0] for key in ftout}
            return ftout[0]
        except IndexError:
            msg = "Cannot find any valid frametypes for channel(s)"
            raise ValueError(msg) from None


@_select_gwdatafind_mod
def find_types(
    observatory: str,
    match: str | re.Pattern | None = None,
    trend: str | None = None,
    **kwargs,
) -> list[str]:
    """Find the available data types for a given observatory.

    Parameters
    ----------
    observatory : `str`
        The observatory for which to search, as a single character,
        e.g ``"G"`` for GEO-600.

    match : `str`, `re.Pattern`, optional
        Regular expression to match types against.

    trend : `str`, optional
        A trend type name to prioritise, e.g. ``"m-trend"``.

    kwargs
        Other arguments are passed to `gwdatafind.find_types`.

    Returns
    -------
    types : `list` of `str`
        A list of matching types, sorted by priority
        (*h(t)* datasets first, large datasets last).

    See Also
    --------
    gwdatafind.find_types
        For details of how available types are discovered.
    """
    types = gwdatafind.find_types(
        observatory,
        match=match,
        **kwargs,
    )
    return sorted(
        types,
        key=partial(_type_priority, trend=trend),
    )


@_select_gwdatafind_mod
def find_urls(
    observatory: str,
    frametype: str,
    start: SupportsToGps,
    end: SupportsToGps,
    on_gaps: Literal["error", "ignore", "warn"] = "error",
    **kwargs,
) -> list[str]:
    """Find the URLs of files of a given data type in a GPS interval.

    Parameters
    ----------
    observatory : `str`
        The observatory for which to search, as a single character,
        e.g ``"G"`` for GEO-600.

    frametype : `str`
        Name of dataset to match.

    start : `~gwpy.time.LIGOTimeGPS`, `int`, `str`
        GPS start time of search.
        any input parseable by `~gwpy.time.to_gps` is fine.
        Non-integer GPS start time will be rounded down.

    end : `~gwpy.time.LIGOTimeGPS`, `int`, `str`
        GPS end time of search.
        any input parseable by `~gwpy.time.to_gps` is fine
        Non-integer GPS end time will be rounded up.

    on_gaps : `str`, optional
        Action to take when the requested all or some of the GPS interval
        is not covereed by the dataset, one of:

        - ``'error'``: raise a `RuntimeError` (default)
        - ``'warn'``: emit a warning but return all available URLs
        - ``'ignore'``: return the list of URLs with no warnings

    kwargs
        Other arguments are passed to `gwdatafind.find_urls`.

    See Also
    --------
    gwdatafind.find_urls
        For details of the underlying discovery tool, and supported
        keyword arguments.
    """
    return gwdatafind.find_urls(
        observatory,
        frametype,
        int(to_gps(start)),
        ceil(to_gps(end)),
        on_gaps=on_gaps,
        **kwargs,
    )


@_select_gwdatafind_mod
def find_latest(
    observatory: str,
    frametype: str,
    gpstime: SupportsToGps | None = None,
    *,
    allow_tape: bool = False,
    **kwargs,
) -> str:
    """Find the path of the latest file of a given data type.

    Parameters
    ----------
    observatory : `str`
        The observatory for which to search, as a single character,
        e.g ``"G"`` for GEO-600.

    frametype : `str`
        Name of dataset to match.

    gpstime : `~gwpy.time.LIGOTimeGPS`, `float`, `str`, optional
        GPS time to search for a file,
        any input parseable by `~gwpy.time.to_gps` is fine
        Default (`None`) is to search for the most recent file available.

    allow_tape : `bool`, optional
        If `False` (default) raise `OSError` if the 'latest' URL available
        points to a local path stored on tape (not on spinning disk).

    kwargs
        Other keyword arguments are passed to `gwdatafind.find_latest` or
        `gwdatafind.find_urls`.

    See Also
    --------
    gwdatafind.find_latest
        For details of how the latest URL is discovered and available keyword
        arguments.
        This is called when ``gpstime=None`` is given.

    gwdatafind.find_urls
        For details of how the URLs are discovered when ``gpstime`` is given,
        and available keyword arguments.
    """
    if SINGLE_IFO_OBSERVATORY.match(observatory):
        observatory = observatory[0]
    try:
        if gpstime is not None:
            gpstime = int(to_gps(gpstime))
            path = find_urls(
                observatory,
                frametype,
                gpstime,
                gpstime+1,
                on_gaps="ignore",
                **kwargs,
            )[-1]
        else:
            path = gwdatafind.find_latest(
                observatory,
                frametype,
                on_missing="ignore",
                **kwargs,
            )[-1]
    except (
        IndexError,
        RuntimeError,
    ) as exc:
        msg = f"no files found for {observatory}-{frametype}"
        raise RuntimeError(msg) from exc

    if not allow_tape and on_tape(path):
        msg = (
            f"Latest frame file for {observatory}-{frametype} is on tape "
            f"(pass allow_tape=True to force): {path}"
        )
        raise OSError(msg)
    return path



# Copyright (c) 2019-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""
.. sectionauthor:: Duncan Macleod <duncan.macleod@ligo.org>
.. currentmodule:: gwpy.timeseries

Plotting a spectrogram of all open data for many hours
######################################################

In order to study interferometer performance, it is common in LIGO to plot
all of the data for a day, in order to determine trends, and see data-quality
issues.

This is done for the LIGO-Virgo detector network, with
`up-to-date plots <https://gwosc.org/detector_status/>`__
available from |GWOSC|.

This example demonstrates how to download data segments from GWOSC, then
use those to build a multi-hour spectrogram plot of LIGO-Livingston strain
data.
"""

# %%
# .. currentmodule:: gwpy.segments
#
# Getting the segments
# --------------------
#
# First, we need to fetch the Open Data timeline segments from GWOSC.
# To do that we can call the :meth:`DataQualityFlag.fetch_open_data` method
# using ``'H1_DATA'`` as the flag (for an explanation of what this means,
# read up on `The S6 Data Release <https://gwosc.org/S6/>`__).

from gwpy.segments import DataQualityFlag
l1segs = DataQualityFlag.fetch_open_data(
    "L1_DATA",
    "Aug 17 2017 08:00",
    "Aug 17 2017 16:00",
)

# %%
# For sanity, lets plot these segments:

splot = l1segs.plot(
    figsize=[12, 3],
    epoch="August 17 2017",
)
splot.show()
splot.close()  # hide

# %%
# We see that the LIGO Hanford Observatory detector was operating for the
# majority of the day, with a few outages of ~30 minutes or so.

# %%
# We can use the :func:`abs` function to display the total amount of time
# spent taking data:

print(abs(l1segs.active))

# %%
# .. currentmodule:: gwpy.timeseries
#
# Working with strain data
# ------------------------
#
# Now, we can loop through the active segments of ``'L1_DATA'`` and fetch the
# strain `TimeSeries` for each segment, calculating a
# :class:`~gwpy.spectrogram.Spectrogram` for each segment.

from gwpy.timeseries import TimeSeries
spectrograms = []
for start, end in l1segs.active:
    l1strain = TimeSeries.get(
        "L1",
        start,
        end,
    )
    specgram = l1strain.spectrogram(30, fftlength=4) ** (1/2.)
    spectrograms.append(specgram)

# %%
# Finally, we can build a :meth:`~gwpy.spectrogram.Spectrogram.plot`:

# Create an empty plot with a single set of Axes
from gwpy.plot import Plot
plot = Plot(figsize=(12, 6))
ax = plot.gca()
# add each spectrogram to the Axes
for specgram in spectrograms:
    ax.imshow(specgram)
# finalise the plot metadata
ax.set_xscale("auto-gps", epoch="Aug 17 2017")
ax.set_ylim(20, 2000)
ax.set_yscale("log")
ax.set_ylabel("Frequency [Hz]")
ax.set_title("LIGO-Livingston strain data")
ax.colorbar(
    cmap="viridis",
    norm="log",
    clim=(5e-24, 1e-21),
    label=r"Strain noise [1/$\sqrt{\mathrm{Hz}}$]",
)
# add the segments as a 'state' indicator along the bottom
plot.add_segments_bar(l1segs)
plot.show()



# Copyright (c) 2020-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.ligolw`."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy
import pytest
from astropy.table import (
    Table,
    vstack,
)
from numpy.testing import (
    assert_allclose,
    assert_array_equal,
)

from ...testing import utils
from ...testing.errors import pytest_skip_network_error
from .. import EventTable
from .utils import random_table

if TYPE_CHECKING:
    from igwn_ligolw.lsctables import SnglBurstTable

TEST_XML_PATH = utils.TEST_DATA_PATH / "H1-LDAS_STRAIN-968654552-10.xml.gz"
REMOTE_XML_FILE = (
    "https://gitlab.com/gwpy/gwpy/-/raw/v3.0.10/gwpy/testing/data/"
    + TEST_XML_PATH.name
)


@pytest.fixture
def table() -> Table:
    """Create a table to use in testing."""
    return random_table(
        names=["peak_time", "peak_time_ns", "snr", "central_freq"],
        dtypes=["i4", "i4", "f4", "f4"],
        length=100,
    )


@pytest.fixture
def llwtable() -> SnglBurstTable:
    """Create a LIGO_LW table to use in testing."""
    from igwn_ligolw.lsctables import SnglBurstTable
    llwtab = SnglBurstTable.new(columns=["peak_frequency", "snr"])
    for i in range(10):
        row = llwtab.RowType()
        row.peak_frequency = float(i)
        row.snr = float(i)
        llwtab.append(row)
    return llwtab


# -- conversions ---------------------

@pytest.mark.requires("igwn_ligolw.lsctables")
def test_to_astropy_table(llwtable):
    """Test converting a LIGO_LW table to an Astropy Table."""
    tab = EventTable(llwtable)
    assert set(tab.colnames) == {"peak_frequency", "snr"}
    assert_array_equal(tab["snr"], llwtable.getColumnByName("snr"))


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_to_astropy_table_rename(llwtable):
    """Test converting a LIGO_LW table to an Astropy Table with renaming."""
    tab = EventTable(llwtable, rename={"peak_frequency": "frequency"})
    assert set(tab.colnames) == {"frequency", "snr"}
    assert_array_equal(
        tab["frequency"],
        llwtable.getColumnByName("peak_frequency"),
    )


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_to_astropy_table_empty():
    """Test converting an empty LIGO_LW table to an Astropy Table."""
    from igwn_ligolw.lsctables import SnglBurstTable
    llwtable = SnglBurstTable.new(
        columns=["peak_time", "peak_time_ns", "ifo"],
    )
    tab = EventTable(llwtable, columns=["peak", "ifo"])
    assert set(tab.colnames) == {"peak", "ifo"}
    assert tab["peak"].dtype.type is numpy.object_
    assert tab["ifo"].dtype.type is numpy.str_


# -- i/o -----------------------------

@pytest.mark.parametrize("ext", ["xml", "xml.gz"])
@pytest.mark.parametrize("table_class", [Table, EventTable])
@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_write_ligolw(table, ext, table_class):
    """Test `Table.read` and `Table.write` with ``format='ligolw'``."""
    table = table_class(table)
    utils.test_read_write(
        table,
        "ligolw",
        ext,
        write_kw={
            "tablename": "sngl_burst",
        },
        autoidentify=False,
        assert_equal=utils.assert_table_equal,
        assert_kw={
            "almost_equal": True,
        },
    )


@pytest.mark.parametrize("use_numpy_dtypes", [False, True])
@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_write_ligolw_types(use_numpy_dtypes):
    """Test `Table.read(format='ligolw')` with ``use_numpy_dtypes``."""
    t2 = Table.read(
        TEST_XML_PATH,
        tablename="sngl_burst",
        columns=["peak"],
        use_numpy_dtypes=use_numpy_dtypes,
    )
    peak = t2["peak"][0]
    if use_numpy_dtypes:
        assert isinstance(peak, float)
    else:
        from igwn_ligolw.lsctables import LIGOTimeGPS as LigolwGps
        assert isinstance(peak, LigolwGps)


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_ligolw_multiple():
    """Test `Table.read(format='ligolw')` with multiple files."""
    t = EventTable.read(
        TEST_XML_PATH,
        tablename="sngl_burst",
    )
    t2 = EventTable.read(
        [TEST_XML_PATH, TEST_XML_PATH],
        tablename="sngl_burst",
    )
    utils.assert_table_equal(t2, vstack((t, t)))


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_write_ligolw_overwrite(table, tmp_path):
    """Test `Table.write(format='ligolw', overwrite=True)` overwrites a table."""
    # write the table once
    xml = tmp_path / "test.xml"
    table.write(xml, format="ligolw", tablename="sngl_burst")

    # check that writing again to the same file fails
    with pytest.raises(
        OSError,
        match=f"^File exists: {xml}$",
    ):
        table.write(xml, format="ligolw", tablename="sngl_burst")

    # but overwrite=True forces the new write
    table.write(
        xml,
        format="ligolw",
        tablename="sngl_burst",
        overwrite=True,
    )

    # check that the table is the same
    utils.assert_table_equal(
        type(table).read(xml, format="ligolw"),
        table,
        almost_equal=True,
    )


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_write_ligolw_append(table, tmp_path):
    """Test `Table.write(format='ligolw', append=True)` extends a table."""
    # write the table once
    xml = tmp_path / "test.xml"
    table.write(xml, format="ligolw", tablename="sngl_burst")

    # test that append=True extends the table in-place
    table.write(xml, format="ligolw", tablename="sngl_burst", append=True)
    utils.assert_table_equal(
        type(table).read(xml, format="ligolw"),
        vstack((table, table)),
        almost_equal=True,
    )


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_write_ligolw_append_multiple_tables(table, tmp_path):
    """Test `Table.write(format='ligolw', append=True)` with different tables."""
    # write the first table
    xml = tmp_path / "test.xml"
    table.write(xml, format="ligolw", tablename="sngl_burst")

    # write another table into the same file
    insp = random_table(
        names=["end_time", "snr", "chisq_dof"],
        length=10,
    )
    insp.write(
        xml,
        format="ligolw",
        tablename="sngl_inspiral",
        append=True,
    )

    # check that we can get back the first, but not have to specify tablename
    with pytest.raises(
        ValueError,
        match="Multiple tables found in LIGO_LW document",
    ):
        type(table).read(xml, format="ligolw")
    utils.assert_table_equal(
        type(table).read(xml, format="ligolw", tablename="sngl_burst"),
        table,
        almost_equal=True,
    )


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_write_ligolw_property_columns(tmp_path):
    """Test reading/writing LIGO_LW files handles property columns properlty.

    As below the 'peak' column is a `gpsproperty` made by dynamically combining
    the ``'peak_time'`` and ``'peak_time_ns'`` columns.
    This test checks that the writer handles this silently, and that the reader
    can handle it.
    """
    # create table that uses a property column ('peak')
    table = random_table(
        names=["peak", "snr", "central_freq"],
        dtypes=["f8", "f4", "f4"],
        length=100,
    )

    # write table to file
    xml = tmp_path / "test.xml"
    table.write(
        xml,
        format="ligolw",
        tablename="sngl_burst",
    )

    # read it back and check that the gpsproperty was unpacked properly
    t2 = type(table).read(
        xml,
        format="ligolw",
        tablename="sngl_burst",
    )
    assert set(t2.colnames) == {
        "peak_time",
        "peak_time_ns",
        "snr",
        "central_freq",
    }
    assert_allclose(
        t2["peak_time"] + t2["peak_time_ns"] * 1e-9,
        table["peak"],
    )

    # test that reading property columns directly works.
    t3 = type(table).read(
        xml,
        format="ligolw",
        tablename="sngl_burst",
        columns=["peak", "peak_time"],
    )
    assert set(t3.colnames) == {"peak", "peak_time"}
    assert_array_equal(
        t3["peak"],
        table["peak"],
    )


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_ligolw_get_as_exclude(tmp_path):
    """Test that reading LIGO_LW files handles columns whose names don't
    correspond to the `get_{name}` method on the parent table.

    E.g. the ``'time_slide'`` table has a column 'time_slide_id' and a
    method `~igwn_ligolw.lsctables.TimeSlideTable.get_time_slide_id` that have
    nothing to do with each other.
    """  # noqa: D205
    # create a time_slide table with a 'time_slide_id' column
    table = Table(
        rows=[
            ("H1", 0.0, 4, 0),
            ("L1", 0.62831, 4, 0),
            ("V1", 0.31415, 4, 0),
        ],
        names=("instrument", "offset", "process_id", "time_slide_id"),
    )

    # write it
    xml = tmp_path / "test.xml"
    table.write(
        xml,
        format="ligolw",
        tablename="time_slide",
    )

    # read it back and check that the `time_slide_id` column is read properly
    t2 = type(table).read(
        xml,
        tablename="time_slide",
    )
    t2.sort("instrument")
    utils.assert_table_equal(t2, table)


@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_process_table():
    """Regression test against gwpy/gwpy#1367."""
    from igwn_ligolw.lsctables import ProcessTable
    llwtable = ProcessTable.new(
        columns=["ifos", "username"],
    )
    llwtable.appendRow(instruments=("G1", "H1"), username="testuser")
    llwtable.appendRow(instruments=("H1", "L1"), username="testuser")
    tab = EventTable(llwtable)
    assert len(tab) == 2
    assert tab[0]["ifos"] == "G1,H1"


@pytest_skip_network_error
@pytest.mark.requires("igwn_ligolw.lsctables")
def test_read_remote_file():
    """Test that we can read remote files over HTTP."""
    tab = EventTable.read(
        REMOTE_XML_FILE,
        cache=False,
        format="ligolw",
        tablename="sngl_burst",
    )
    assert len(tab) == 2052
    assert tab[0]["snr"] == pytest.approx(0.69409615)



# Copyright (c) 2024-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.tools.rds`."""

from ...testing.errors import pytest_skip_network_error
from ...timeseries import TimeSeries
from ..rds import main as gwpy_rds


@pytest_skip_network_error
def test_rds(tmp_path):
    """Test the ``gwpy-rds`` tool."""
    ifo = "H1"
    outfile = tmp_path / "test.h5"
    gwpy_rds([
        "1126259460",
        "1126259464",
        ifo,
        "-o", str(outfile),
        "-O", "format=hdf5",
        "-O", "sample_rate=4096",
        "-O", "version=4",
    ])
    data = TimeSeries.read(outfile, ifo)
    assert data.name == "H1:Strain"
    assert data.span == (1126259460, 1126259464)



# Copyright (c) 2023-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.utils.sphinx.gallery`."""

from __future__ import annotations

import importlib.metadata
from configparser import ConfigParser
from typing import TYPE_CHECKING

import pytest

from ..sphinx import gallery

if TYPE_CHECKING:
    from pathlib import Path

# load the default entry point
GWPY_PLOT, = importlib.metadata.entry_points(name="gwpy-plot")

# example configuration
TEST_CONFIG = {
    "test-1": {
        "title": "Title 1",
        "description": "Description 1",
        "argv": "one two three --four five",
    },
}


@pytest.fixture
def ini() -> ConfigParser:
    """Format ``TEST_CONFIG`` as a `configparser.ConfigParser`."""
    ini = ConfigParser()
    ini.update(TEST_CONFIG)
    return ini


@pytest.fixture
def inifile(tmp_path, ini) -> Path:
    """Write ``ini`` to a temporary file and return the path."""
    inifile = tmp_path / "examples.ini"
    with inifile.open("w") as conf:
        ini.write(conf)
    return inifile


def test_render_entry_point_examples(tmp_path, inifile):
    """Test `render_entry_point_examples`."""
    gallery.render_entry_point_examples(
        inifile,
        tmp_path,
        entry_point=GWPY_PLOT.name,
        filename_prefix="test_",
    )

    # check gallery header
    assert (tmp_path / "GALLERY_HEADER.rst").read_text().startswith(
        "Click on a thumbnail",
    )

    test1py = (tmp_path / "test_test-1.py").read_text()

    # check title and description
    assert """
Title 1
#######

Description 1""" in test1py

    # check python function execution
    assert f"""
from {GWPY_PLOT.value.replace(':', ' import ')}
main([
    'one', 'two', 'three',
    '--four', 'five',
])""" in test1py
